# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/05_token_classification.ipynb (unless otherwise specified).

__all__ = ['logger', 'TokenClassificationResult', 'TransformersTokenTagger', 'FlairTokenTagger', 'EasyTokenTagger']

# Cell
import logging
from typing import List, Dict, Union
from collections import defaultdict, OrderedDict

import numpy as np

import torch
from torch.utils.data import TensorDataset

from flair.data import Sentence
from flair.models import SequenceTagger
from transformers import (
    AutoTokenizer,
    AutoModelForTokenClassification,
    PreTrainedTokenizer,
    PreTrainedModel,
    BertForSequenceClassification,
    XLNetForSequenceClassification,
    AlbertForSequenceClassification,
)

from ..result import DetailLevel

from ..model import AdaptiveModel, DataLoader
from ..model_hub import HFModelResult, FlairModelResult, FlairModelHub, HFModelHub

from fastai.torch_core import to_detach, apply, to_device

from fastcore.basics import Self, risinstance
from fastcore.xtras import Path

# Cell
logger = logging.getLogger(__name__)

# Cell
class TokenClassificationResult:
    "A result class for Token Tagging tasks"
    def __init__(
        self,
        inputs:str, # Original text inputs
        tokenized_inputs:torch.tensor, # Tokenized inputs
        tagged_entities:dict
    ):
        self.inputs = inputs
        self.tokenized_inputs = tokenized_inputs
        self.tagged_entities = tagged_entities

    def to_dict(
        self,
        detail_level:DetailLevel=DetailLevel.Low # A level of detail to return
    ) -> dict:
        "Convert `self` to a dictionary"
        o = OrderedDict()
        o.update({
            'tags': self.tagged_entities
        })
        if detail_level == 'medium' or detail_level == 'high':
            o.update({
                'inputs':self.inputs,
                'tokenized_inputs':self.tokenized_inputs
            })
        if detail_level != 'high':
            try:
                for tag in o['tags'][0]:
                    tag.pop('offsets')
            except:
                return o
        return o

# Cell
class TransformersTokenTagger(AdaptiveModel):
    "Adaptive model for Transformer's Token Tagger Model"
    def __init__(
        self,
        tokenizer: PreTrainedTokenizer, # A tokenizer object from Huggingface's transformers (TODO) and tokenizers
        model: PreTrainedModel # A transformers token tagger model
    ):
        # Load up model and tokenizer
        self.tokenizer = tokenizer
        super().__init__()

        # Sets the internal model
        self.set_model(model)

    @classmethod
    def load(
        cls,
        model_name_or_path: str # A key string of one of Transformer's pre-trained Token Tagger Model or a `HFModelResult`
    ) -> AdaptiveModel:
        "Class method for loading and constructing this tagger"
        if isinstance(model_name_or_path, HFModelResult): model_name_or_path = model_name_or_path.name
        tokenizer = AutoTokenizer.from_pretrained(model_name_or_path)
        model = AutoModelForTokenClassification.from_pretrained(model_name_or_path)
        tagger = cls(tokenizer, model)
        return tagger

    def predict(
        self,
        text: Union[List[str], str], # Sentences to run inference on
        mini_batch_size: int = 32, # Mini batch size
        grouped_entities: bool = True, # Return whole entity span strings
        detail_level:DetailLevel = DetailLevel.Low, # A level of detail to return
        **kwargs, # Optional arguments for the Transformers tagger
    ) -> List[List[Dict]]: # Returns a list of lists of tagged entities
        "Predict method for running inference using the pre-trained token tagger model"
        if isinstance(text, str):
            text = [text]
        results: List[Dict] = []

        dataset = self._tokenize(text)
        dl = DataLoader(dataset, batch_size=mini_batch_size)

        logger.info(f'Running prediction on {len(dataset)} text sequences')
        logger.info(f'Batch size = {mini_batch_size}')

        outputs,_ = super().get_preds(dl=dl)

        inputs = apply(to_device, [b for b in dl], device='cpu')
        inputs = torch.cat(*inputs)
        inputs = apply(Self.numpy(), inputs)

        outputs = torch.cat([o['logits'] for o in outputs])
        outputs = apply(to_detach, outputs, cpu=True)
        outputs = apply(Self.numpy(), outputs)

        # Iterate through batch for tagged token predictions
        for idx, pred in enumerate(outputs):
            entities = pred
            input_ids = inputs[idx]
            tagged_entities = self._generate_tagged_entities(
                entities=entities,
                input_ids=input_ids,
                grouped_entities=grouped_entities
            )
            results += tagged_entities

        results = TokenClassificationResult(text, inputs, results)

        return results.to_dict(detail_level) if detail_level is not None else detail_level

    def _tokenize(
        self, sentences: Union[List[Sentence], Sentence, List[str], str]
    ) -> TensorDataset:
        "Batch tokenizes text and produces a `TensorDataset` with them"

        tokenized_text = self.tokenizer.batch_encode_plus(
            sentences,
            return_tensors="pt",
            max_length=None,
        )

        # Bart, XLM, DistilBERT, RoBERTa, and XLM-RoBERTa don't use token_type_ids
        if isinstance(
            self.model,
            (
                BertForSequenceClassification,
                XLNetForSequenceClassification,
                AlbertForSequenceClassification,
            ),
        ):
            dataset = TensorDataset(
                tokenized_text["input_ids"],
                tokenized_text["attention_mask"],
                tokenized_text["token_type_ids"],
            )
        else:
            dataset = TensorDataset(
                tokenized_text["input_ids"], tokenized_text["attention_mask"]
            )

        return dataset

    # `_group_entites` and `_generate_tagged_entities` modified from pipeline code snippet from Transformers
    def _group_entities(
        self, entities: List[dict], idx_start: int, idx_end: int
    ) -> Dict:
        """Returns grouped entities"""
        # Get the last entity in the entity group
        entity = entities[-1]["entity"]
        scores = np.mean([entity["score"] for entity in entities])
        tokens = [entity["word"] for entity in entities]

        entity_group = {
            "entity": entity,
            "score": np.mean(scores),
            "word": self.tokenizer.convert_tokens_to_string(tokens),
            "offsets": (idx_start, idx_end),
        }
        return entity_group

    def _generate_tagged_entities(
        self, entities: np.ndarray, input_ids: np.ndarray, grouped_entities: bool = True
    ) -> List[Dict]:
        """Generate full list of entities given tagged token predictions and input_ids"""

        score = np.exp(entities) / np.exp(entities).sum(-1, keepdims=True)
        labels_idx = score.argmax(axis=-1)

        answers = []
        entities = []
        entity_groups = []
        entity_group_disagg = []
        # Filter to labels not in `self.ignore_labels`
        filtered_labels_idx = [
            (idx, label_idx)
            for idx, label_idx in enumerate(labels_idx)
            if self.model.config.id2label[label_idx] not in ["O"]
        ]

        for idx, label_idx in filtered_labels_idx:
            # print(tokenizer.convert_ids_to_tokens(int(input_ids[idx])))
            entity = {
                "word": self.tokenizer.convert_ids_to_tokens(int(input_ids[idx])),
                "score": score[idx][label_idx].item(),
                "entity": self.model.config.id2label[label_idx],
                "index": idx,
            }
            last_idx, _ = filtered_labels_idx[-1]
            if grouped_entities:
                if not entity_group_disagg:
                    entity_group_disagg += [entity]
                    if idx == last_idx:
                        entity_groups += [
                            self._group_entities(
                                entity_group_disagg, idx - len(entity_group_disagg), idx
                            )
                        ]
                    continue

                # If the current entity is similar and adjacent to the previous entity, append it to the disaggregated entity group
                if (
                    entity["entity"] == entity_group_disagg[-1]["entity"]
                    and entity["index"] == entity_group_disagg[-1]["index"] + 1
                ):
                    entity_group_disagg += [entity]
                    # Group the entities at the last entity
                    if idx == last_idx:
                        entity_groups += [
                            self._group_entities(
                                entity_group_disagg, idx - len(entity_group_disagg), idx
                            )
                        ]
                # If the current entity is different from the previous entity, aggregate the disaggregated entity group
                else:
                    entity_groups += [
                        self._group_entities(
                            entity_group_disagg,
                            entity_group_disagg[-1]["index"] - len(entity_group_disagg),
                            entity_group_disagg[-1]["index"],
                        )
                    ]
                    entity_group_disagg = [entity]

            entities += [entity]

        # Append
        if grouped_entities:
            answers += [entity_groups]
        else:
            answers += [entities]

        return answers

# Cell
class FlairTokenTagger(AdaptiveModel):
    "Adaptive Model for Flair's Token Tagger"

    def __init__(
        self,
        model_name_or_path: str # A key string of one of Flair's pre-trained Token tagger Model, [link](https://huggingface.co/models?filter=flair)
    ):
        self.tagger = SequenceTagger.load(model_name_or_path)

    @classmethod
    def load(
        cls,
        model_name_or_path: str # A key string of one of Flair's pre-trained Token tagger Model, [link](https://huggingface.co/models?filter=flair)
    ) -> AdaptiveModel:
        "Class method for loading a constructing this tagger"
        tagger = cls(model_name_or_path)
        return tagger

    def predict(
        self,
        text: Union[List[Sentence], Sentence, List[str], str], # Sentences to run inference on
        mini_batch_size: int = 32, # Mini batch size
        **kwargs, # Optional arguments for the Flair tagger
    ) -> List[Sentence]: # A list of predicted sentences
        "Predict method for running inference using the pre-trained token tagger model"

        if isinstance(text, (Sentence, str)):
            text = [text]
        if isinstance(text[0], str):
            text = [Sentence(s) for s in text]
        self.tagger.predict(
            sentences=text,
            mini_batch_size=mini_batch_size,
            **kwargs,
        )
        return text

# Cell
class EasyTokenTagger:
    "Token level classification models"

    def __init__(self):
        self.token_taggers: Dict[AdaptiveModel] = defaultdict(bool)

    def tag_text(
        self,
        text: Union[List[Sentence], Sentence, List[str], str], # Text input, it can be a string or any of Flair's `Sentence` input formats
        model_name_or_path: Union[str, FlairModelResult, HFModelResult] = "ner-ontonotes", # The hosted model name key or model path
        mini_batch_size: int = 32, # The mini batch size for running inference
        detail_level:DetailLevel = DetailLevel.Low, # The level fo detail to return on a TransformerTagger
        **kwargs, # Keyword arguments for Flair's `SequenceTagger.predict()` method
    ) -> List[Sentence]: # A list of Flair's `Sentence`'s
        "Tags tokens with labels the token classification models have been trained on"
        # Load Sequence Tagger Model and Pytorch Module into tagger dict
        name = getattr(model_name_or_path, 'name', model_name_or_path)
        if not self.token_taggers[name]:
            """
            self.token_taggers[model_name_or_path] = SequenceTagger.load(
                model_name_or_path
            )
            """
            if risinstance([FlairModelResult, HFModelResult], model_name_or_path):
                try:
                    self.token_taggers[name] = FlairTokenTagger.load(name)
                except:
                    self.token_taggers[name] = TransformersTokenTagger.load(name)
            elif risinstance([str, Path], model_name_or_path) and (Path(model_name_or_path).exists() and Path(model_name_or_path).is_dir()):
                # Load in previously existing model
                try:
                    self.token_taggers[name] = FlairTokenTagger.load(name)
                except:
                    self.token_taggers[name] = TransformersTokenTagger.load(name)
            else:
                _flair_hub = FlairModelHub()
                _hf_hub = HFModelHub()
                res = _flair_hub.search_model_by_name(name, user_uploaded=True)
                if len(res) < 1:
                    # No models found
                    res = _hf_hub.search_model_by_name(name, user_uploaded=True)
                    if len(res) < 1:
                        logger.info("Not a valid `model_name_or_path` param")
                        return [Sentence('')]
                    else:
                        res[0].name.replace('flairNLP', 'flair')
                        self.token_taggers[res[0].name] = TransformersTokenTagger.load(res[0].name)
                        name = res[0].name

                else:
                    name = res[0].name.replace('flairNLP/', '')
                    self.token_taggers[name] = FlairTokenTagger.load(name) # Returning the first should always be the non-fast option

        tagger = self.token_taggers[name]
        if isinstance(tagger, TransformersTokenTagger):
            return tagger.predict(
                text=text,
                mini_batch_size=mini_batch_size,
                **kwargs
            )
        else:
            return tagger.predict(
                text=text,
                mini_batch_size=mini_batch_size,
                **kwargs,
            )

    def tag_all(
        self,
        text: Union[List[Sentence], Sentence, List[str], str], # Text input, it can be a string or any of Flair's `Sentence` input formats
        mini_batch_size: int = 32, # The mini batch size for running inference
        detail_level:DetailLevel = DetailLevel.Low, # The level of detail for a TransformersTagger to return
        **kwargs, # Keyword arguments for Flair's `SequenceTagger.predict()` method
    ) -> List[Sentence]: # A list of Flair's `Sentence`'s
        "Tags tokens with all labels from all token classification models"
        if len(self.token_taggers) == 0:
            print("No token classification models loaded...")
            return Sentence()
        sentences = text
        for tagger_name in self.token_taggers.keys():
            sentences = self.tag_text(
                sentences,
                model_name_or_path=tagger_name,
                mini_batch_size=mini_batch_size,
                **kwargs,
            )
        return sentences


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/08_translation.ipynb (unless otherwise specified).

__all__ = ['logger', 'TranslationResult', 'TransformersTranslator', 'EasyTranslator']

# Cell
import logging
from typing import List, Dict, Union
from collections import defaultdict, OrderedDict

import torch
from torch.utils.data import TensorDataset, DataLoader

from transformers import (
    AutoTokenizer,
    AutoModelForSeq2SeqLM,
    PreTrainedTokenizer,
    PreTrainedModel,
    T5ForConditionalGeneration,
)

from ..model import AdaptiveModel
from ..callback import GeneratorCallback

from fastai.torch_core import apply, to_device

from fastcore.basics import Self
from ..model_hub import HFModelResult, FlairModelResult, HFModelHub, FlairModelHub
from ..result import DetailLevel

# Cell
logger = logging.getLogger(__name__)

# Cell
class TranslationResult:
    "A basic result class for Translation problems"
    def __init__(
        self,
        inputs:List[str], # A list of input string sentences
        input_lang:str, # A input language
        output_lang:str, # An output language
        translations:List[str] # A list of the translated sentences
    ):
        self.inputs = inputs
        self.input_lang = input_lang
        self.output_lang = output_lang
        self.translations = translations

    def to_dict(
        self,
        detail_level:DetailLevel=DetailLevel.Low # A detail level to return
    ):
        "Convert `self` to a filtered dictionary"
        o = OrderedDict()
        o.update(
            {'translations':self.translations,}
        )
        if detail_level == 'medium' or detail_level == 'high':
            o.update({
                'inputs': self.inputs,
                'input_language': self.input_lang,
                'output_language': self.output_lang
            })
        if detail_level == 'high':
            print("Warning: There is no difference between `medium` and `high` for Translation tasks")
        return o

# Cell
class TransformersTranslator(AdaptiveModel):
    "Adaptive model for Transformer's Conditional Generation or Language Models (Transformer's T5 and Bart conditional generation models have a language modeling head)"

    def __init__(
        self,
        tokenizer: PreTrainedTokenizer, # A tokenizer object from Huggingface's transformers (TODO)and tokenizers
        model: PreTrainedModel # A transformers Conditional Generation (Bart or T5) or Language model
    ):
        # Load up model and tokenizer
        self.tokenizer = tokenizer
        super().__init__()

        # Sets internal model
        self.set_model(model)

        # Set inputs to come in as `dict`
        super().set_as_dict(True)

    @classmethod
    def load(
        cls,
        model_name_or_path: str # A key string of one of Transformer's pre-trained translator Model
    ) -> AdaptiveModel:
        "Class method for loading and constructing this classifier"
        tokenizer = AutoTokenizer.from_pretrained(model_name_or_path)
        model = AutoModelForSeq2SeqLM.from_pretrained(model_name_or_path)
        translator = cls(tokenizer, model)
        return translator

    def predict(
        self,
        text: Union[List[str], str], # Sentences to run inference on
        t5_prefix: str = 'translate English to German', # The pre-appended prefix for the specificied task. Only in use for T5-type models.
        mini_batch_size: int = 32,  # Mini batch size
        num_beams: int = 1, # Number of beams for beam search. Must be between 1 and infinity. 1 means no beam search
        min_length: int = 0, # The min length of the sequence to be generated
        max_length: int = 128, # The max length of the sequence to be generated. Between min_length and infinity
        early_stopping: bool = True, # If set to `True` beam search is stopped when at least num_beams sentences finished per batch
        detail_level:DetailLevel = DetailLevel.Low, # The level of detail to return
        **kwargs, # Optional arguments for the Transformers `PreTrainedModel.generate()` method
    ) -> List[str]: # A list of translated sentences
        "Predict method for running inference using the pre-trained sequence classifier model. Keyword arguments for parameters of the method `Transformers.PreTrainedModel.generate()` can be used as well"

        # Make all inputs lists
        if isinstance(text, str):
            text = [text]

        # T5 requires 'translate: ' precursor text for pre-trained translator
        if isinstance(self.model, T5ForConditionalGeneration):
            text = [f'{t5_prefix}: {t}' for t in text]

        dataset = self._tokenize(text)
        dl = DataLoader(dataset, batch_size=mini_batch_size)
        translations = []

        logger.info(f'Running translator on {len(dataset)} text sequences')
        logger.info(f'Batch size = {mini_batch_size}')

        cb = GeneratorCallback(num_beams, min_length, max_length, early_stopping, **kwargs)

        preds,_ = super().get_preds(dl=dl, cbs=[cb])

        preds = apply(Self.squeeze(0), preds)

        for o in preds:
            translations.append(
                [
                    self.tokenizer.decode(
                        o,
                        skip_special_tokens=True,
                        clean_up_tokenization_spaces=False,
                    )
                ].pop()
            )

        languages = t5_prefix.strip('translate ').split(' to ')

        res = TranslationResult(text, *languages, translations)

        return res if detail_level is None else res.to_dict(detail_level)

    def _tokenize(self, text: Union[List[str], str]) -> TensorDataset:
        """ Batch tokenizes text and produces a `TensorDataset` with text """

        tokenized_text = self.tokenizer.batch_encode_plus(
            text,
            return_tensors='pt',
            max_length=512,
            padding='max_length',
            add_special_tokens=True,
        )

        # Bart doesn't use `token_type_ids`
        dataset = TensorDataset(
            tokenized_text['input_ids'],
            tokenized_text['attention_mask'],
        )
        return dataset

# Cell
class EasyTranslator:
    "Translation Module"

    def __init__(self):
        self.translators: Dict[AdaptiveModel] = defaultdict(bool)

    def translate(
        self,
        text: Union[List[str], str], # Sentences to run inference on
        model_name_or_path: str = "t5-small", # A model id or path to a pre-trained model repository or custom trained model directory
        t5_prefix: str = "translate English to German", # The pre-appended prefix for the specificied task. Only in use for T5-type models
        detail_level=DetailLevel.Low, # The level of detail to return
        mini_batch_size: int = 32, # Mini batch size
        num_beams: int = 1, # Number of beams for beam search. Must be between 1 and infinity. 1 means no beam search
        min_length: int = 0, # The min length of the sequence to be generated
        max_length: int = 128, # The max length of the sequence to be generated. Between min_length and infinity
        early_stopping: bool = True, # If set to `True` beam search is stopped when at least num_beams sentences finished per batch
        **kwargs, # Optional arguments for the Transformers `PreTrainedModel.generate()` method
    ) -> List[str]:
        "Predict method for running inference using the pre-trained sequence classifier model. Keyword arguments for parameters of the method `Transformers.PreTrainedModel.generate()` can be used as well."
        name = getattr(model_name_or_path, 'name', model_name_or_path)
        if not self.translators[name]:
            self.translators[name] = TransformersTranslator.load(
                name
            )

        translator = self.translators[name]
        return translator.predict(
            text=text,
            t5_prefix=t5_prefix,
            mini_batch_size=mini_batch_size,
            num_beams=num_beams,
            min_length=min_length,
            max_length=max_length,
            early_stopping=early_stopping,
            **kwargs,
        )


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/00_result.ipynb (unless otherwise specified).

__all__ = ['DetailLevel', 'SentenceResult']

# Cell
from flair.data import Sentence
from typing import List
import torch

#nbdev_comment _all_ = ['DetailLevel']

# Cell
from fastcore.basics import mk_class

mk_class('DetailLevel', **{o:o.lower() for o in 'High,Medium,Low'.split(',')},
         doc="All possible naming conventions for DetailLevel with typo-proofing")

# Cell
class SentenceResult:
    """
    A result class designed around Flair `Sentences`

    The base class has `tokenized_inputs` and `inputs`
    """
    def __init__(
        self,
        sentences:List[Sentence] # A list of `Sentence` objects
    ):
        self._sentences = sentences

    @property
    def tokenized_inputs(
        self
    ) -> List[str]:
        "The original tokenized inputs"
        return [s.to_tokenized_string() for s in self._sentences]

    @property
    def inputs(
        self
    ) -> List[str]:
        "The original text inputs"
        return [s.to_original_text() for s in self._sentences]


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/10_question_answering.ipynb (unless otherwise specified).

__all__ = ['logger', 'QACallback', 'QAResult', 'TransformersQuestionAnswering', 'EasyQuestionAnswering']

# Cell
import logging
from torch import tensor
from typing import Tuple, List, Union, Dict
from collections import OrderedDict, defaultdict
from tqdm import tqdm

import torch
from torch.utils.data import DataLoader
from transformers import (
    AutoTokenizer,
    AutoModelForQuestionAnswering,
    XLNetForQuestionAnswering,
    XLMForQuestionAnswering,
    CamembertForQuestionAnswering,
    DistilBertForQuestionAnswering,
    RobertaForQuestionAnswering,
    PreTrainedModel,
    PreTrainedTokenizer,
    SquadExample,
    squad_convert_examples_to_features,
)
from transformers.data.processors.squad import SquadResult

from ..model import AdaptiveModel, DataLoader
from ..model_hub import HFModelResult
from .utils import (
    compute_predictions_log_probs,
    compute_predictions_logits,
)

from fastcore.basics import risinstance, nested_attr, Self, patch, listify

from fastai.callback.core import Callback
from fastai.torch_core import apply, to_detach

# Cell
logger = logging.getLogger(__name__)

# Cell
class QACallback(Callback):
    "Basic Question Answering Data Callback"
    order = -2
    _qa_models = [XLMForQuestionAnswering, RobertaForQuestionAnswering, DistilBertForQuestionAnswering]

    def __init__(self, xmodel_instances, features):
        self.xmodel_instances = xmodel_instances
        self.features = features

    def before_batch(self):
        "Adjusts `token_type_ids` if model is in `_qa_models`"
        if risinstance(self._qa_models, self.learn.model): del self.learn.inputs["token_type_ids"]
        if len(self.xb) > 3: self.example_indices = self.xb[3]

        if isinstance(self.learn.model, self.xmodel_instances):
            self.learn.inputs.update({'cls_index': self.xb[4], 'p_mask': self.xb[5]})
            # for lang_id-sensitive xlm models
            if nested_attr(self.learn.model, 'config.lang2id', False):
                # Set language id as 0 for now
                self.learn.inputs.update(
                    {
                        'langs': (
                        torch.ones(self.xb[0].shape, dtype=torch.int64) * 0
                        )
                    }
                )
    def after_pred(self):
        "Generate SquadResults"
        for i, example_index in enumerate(self.example_indices):
            eval_feature = self.features[example_index.item()]
            unique_id = int(eval_feature.unique_id)

            output = [self.pred[output][i] for output in self.pred]
            output = apply(Self.numpy(), to_detach(output))

            if isinstance(self.learn.model, self.xmodel_instances):
                # Some models like the ones in `self.xmodel_instances` use 5 arguments for their predictions
                start_logits = output[0]
                start_top_index = output[1]
                end_logits = output[2]
                end_top_index = output[3]
                cls_logits = output[4]

                self.learn.pred = SquadResult(
                    unique_id,
                    start_logits,
                    end_logits,
                    start_top_index=start_top_index,
                    end_top_index=end_top_index,
                    cls_logits=cls_logits
                )
            else:
                start_logits, end_logits = output
                self.learn.pred = SquadResult(unique_id, start_logits, end_logits)

# Cell
from ..result import DetailLevel

class QAResult:
    """
    A result class designed for Question Answering models
    """
    def __init__(self, examples:List[SquadExample], top_predictions:Union[str, OrderedDict], all_nbest_json:List[OrderedDict]):
        self._examples = examples
        self._top_predictions = top_predictions
        self._all_nbest_json = all_nbest_json

    @property
    def queries(self) -> List[str]:
        """
        The original queries
        """
        return [e.question_text for e in self._examples]

    @property
    def contexts(self) -> List[str]:
        """
        The original context or contexts
        """
        if len(self._examples) < 2:
            return self._examples[0].context_text
        else:
            return [e.context_text for e in self._examples]

    @property
    def probs(self) -> List[List[tensor]]:
        """
        The probabilities returned for each question
        """
        return torch.stack([tensor([o['probability'] for o in self._all_nbest_json[i]]) for i in self._all_nbest_json], dim=0)

    @property
    def best_answer(self) -> List[str]:
        """
        The best answer from each question
        """
        return [self._top_predictions[i] for i in self._top_predictions]

    @property
    def all_answers(self) -> List[List[str]]:
        """
        Every answer ordered by probability from each question
        """
        return tuple([tuple([o['text'] for o in self._all_nbest_json[i]]) for i in self._all_nbest_json])

    def to_dict(self, detail_level:DetailLevel=DetailLevel.Low):
        """
        Return details about `self` at various detail levels
        """
        o = {
                'queries':self.queries,
                'best_answers':self.best_answer,
            }
        if detail_level == 'medium' or detail_level == 'high':
            # Add a dictionary of query and answer and probabilities, also returns contexts
            o['pairings'] = OrderedDict({
                q:(a,p) for (q,a,p) in zip(self.queries, self.all_answers, self.probs)
            })
            o['context'] = self.contexts
        if detail_level == 'high':
            # Add SquadExamples, and n_best_json
            o['squad_example'] = self._examples
            o['n_best_json'] = self._all_nbest_json

        return o

# Cell
class TransformersQuestionAnswering(AdaptiveModel):
    """Adaptive Model for Transformers Question Answering Model

    **Parameters**

    * **tokenizer** - A tokenizer object from Huggingface's transformers (TODO)and tokenizers *
    * **model** - A transformer Question Answering model

    """

    def __init__(self, tokenizer: PreTrainedTokenizer, model: PreTrainedModel):
        # Load up model and tokenizer
        self.tokenizer = tokenizer
        super().__init__()

        # Sets internal model
        self.set_model(model)
        self.xmodel_instances = (XLNetForQuestionAnswering, XLMForQuestionAnswering)

    @classmethod
    def load(cls, model_name_or_path: str) -> AdaptiveModel:
        """Class method for loading and constructing this model

        * **model_name_or_path** - A key string of one of Transformer's pre-trained Question Answering (SQUAD) models
        """
        # QA tokenizers not compatible with fast tokenizers yet
        tokenizer = AutoTokenizer.from_pretrained(model_name_or_path, use_fast=False)
        model = AutoModelForQuestionAnswering.from_pretrained(model_name_or_path)
        qa_model = cls(tokenizer, model)
        return qa_model

    def predict(
        self,
        query: Union[List[str], str],
        context: Union[List[str], str],
        n_best_size: int = 5,
        mini_batch_size: int = 32,
        max_answer_length: int = 10,
        do_lower_case: bool = False,
        version_2_with_negative: bool = False,
        verbose_logging: bool = False,
        null_score_diff_threshold: float = 0.0,
        max_seq_length: int = 512,
        doc_stride: int = 128,
        max_query_length: int = 64,
        **kwargs,
    ) -> Tuple[Tuple[str, List[OrderedDict]], Tuple[OrderedDict, OrderedDict]]:
        """Predict method for running inference using the pre-trained question answering model

        * **query** - String or list of strings that specify the ordered questions corresponding to `context`
        * **context** - String or list of strings that specify the ordered contexts corresponding to `query`
        * **n_best_size** - Number of top n results you want
        * **mini_batch_size** - Mini batch size
        * **max_answer_length** - Maximum token length for answers that are returned
        * **do_lower_case** - Set as `True` if using uncased QA models
        * **version_2_with_negative** - Set as True if using QA model with SQUAD2.0
        * **verbose_logging** - Set True if you want prediction verbose loggings
        * **null_score_diff_threshold** - Threshold for predicting null(no answer) in Squad 2.0 Model.  Default is 0.0.  Raise this if you want fewer null answers
        * **max_seq_length** - Maximum context token length. Check model configs to see max sequence length the model was trained with
        * **doc_stride** - Number of token strides to take when splitting up conext into chunks of size `max_seq_length`
        * **max_query_length** - Maximum token length for queries
        * **&ast;&ast;kwargs**(Optional) - Optional arguments for the Transformers model (mostly for saving evaluations)
        """
        # Make string input consistent as list
        if isinstance(query, str):
            query = [query]
            context = [context]
        assert len(query) == len(context)

        examples = self._mini_squad_processor(query=query, context=context)
        features, dataset = squad_convert_examples_to_features(
            examples,
            self.tokenizer,
            max_seq_length=max_seq_length,
            doc_stride=doc_stride,
            max_query_length=max_query_length,
            is_training=False,
            return_dataset='pt',
            threads=1,
        )
        all_results = []

        dl = DataLoader(dataset, batch_size=mini_batch_size)

        cb = QACallback(self.xmodel_instances, features)

        all_results, _ = super().get_preds(dl=dl, cbs=[cb])

        if isinstance(self.model, self.xmodel_instances):
            start_n_top = (
                self.model.config.start_n_top
                if hasattr(self.model, 'config')
                else self.model.module.config.start_n_top
            )
            end_n_top = (
                self.model.config.end_n_top
                if hasattr(self.model, 'config')
                else self.model.module.config.end_n_top
            )

            answers, n_best = compute_predictions_log_probs(
                examples,
                features,
                all_results,
                n_best_size,
                max_answer_length,
                start_n_top,
                end_n_top,
                version_2_with_negative,
                self.tokenizer,
                verbose_logging,
                **kwargs,
            )

        else:
            answers, n_best = compute_predictions_logits(
                examples,
                features,
                all_results,
                n_best_size,
                max_answer_length,
                do_lower_case,
                verbose_logging,
                version_2_with_negative,
                null_score_diff_threshold,
                self.tokenizer,
                **kwargs,
            )

        return examples, answers, n_best

    def _mini_squad_processor(
        self, query: List[str], context: List[str]
    ) -> List[SquadExample]:
        """Squad data processor to create `SquadExamples`

        * **query** - List of query strings, must be same length as `context`
        * **context** - List of context strings, must be same length as `query`

        """
        assert len(query) == len(context)
        examples = []
        title = 'qa'
        is_impossible = False
        answer_text = None
        start_position_character = None
        answers = ['answer']
        for idx, (q, c) in enumerate(zip(query, context)):
            example = SquadExample(
                qas_id=str(idx),
                question_text=q,
                context_text=c,
                answer_text=answer_text,
                start_position_character=start_position_character,
                title=title,
                is_impossible=is_impossible,
                answers=answers,
            )
            examples.append(example)
        return examples

# Cell
class EasyQuestionAnswering:
    """Question Answering Module

    Usage:

    ```python
    >>> qa = adaptnlp.EasyQuestionAnswering()
    >>> qa.predict_qa(query='What is life?', context='Life is NLP.', n_best_size=5, mini_batch_size=1)
    ```
    """

    def __init__(self):
        self.models: Dict[AdaptiveModel] = defaultdict(bool)

    def predict_qa(
        self,
        query: Union[List[str], str],
        context: Union[List[str], str],
        n_best_size: int = 5,
        mini_batch_size: int = 32,
        model_name_or_path: Union[str, HFModelResult] = 'bert-large-uncased-whole-word-masking-finetuned-squad',
        detail_level = DetailLevel.Low,
        **kwargs,
    ) -> Union[QAResult, dict]:
        """Predicts top_n answer spans of query in regards to context

        * **query** - String or list of strings that specify the ordered questions corresponding to `context`
        * **context** - String or list of strings that specify the ordered contexts corresponding to `query`
        * **n_best_size** - The top n answers returned
        * **mini_batch_size** - Mini batch size for inference
        * **model_name_or_path** - Path to QA model or name of QA model at huggingface.co/models
        * **detail_level** - String or DetailLevel of what amount of information should be returned. If `None` will return `QAResult`
        * **kwargs**(Optional) - Keyword arguments for `AdaptiveModel`s like `TransformersQuestionAnswering`

        **return** - Either a dictionary of results or a QAResult
        """
        name = getattr(model_name_or_path, 'name', model_name_or_path)
        try:
            if not self.models[name]:
                self.models[name] = TransformersQuestionAnswering.load(
                    name
                )
        except OSError:
            logger.info(
                f'{name} not a valid Transformers pre-trained QA model...check path or huggingface.co/models'
            )
            raise ValueError(
                f'{name} is not a valid path or model name from huggingface.co/models'
            )
            return OrderedDict(), [OrderedDict()]

        model = self.models[name]

        examples, top_answer, top_n_answers = model.predict(
            query=query,
            context=context,
            n_best_size=n_best_size,
            mini_batch_size=mini_batch_size,
            **kwargs,
        )

        result = QAResult(examples, top_answer, top_n_answers)

        return result.to_dict(detail_level) if detail_level is not None else result


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/02_model_hub.ipynb (unless otherwise specified).

__all__ = ['HF_TASKS', 'FLAIR_TASKS', 'HFModelResult', 'HFModelHub', 'FLAIR_MODELS', 'FlairModelResult',
           'FlairModelHub']

# Cell
from fastcore.basics import Self, merge
from fastcore.utils import dict2obj, obj2dict, mk_class
from fastai.torch_core import apply
from huggingface_hub.hf_api import ModelInfo, HfApi

from typing import List, Dict

# Internal Cell
_hf_tasks = {
    'FILL_MASK':'fill-mask',
    'QUESTION_ANSWERING':'question-answering',
    'SUMMARIZATION':'summarization',
    'TABLE_QUESTION_ANSWERING':'table-question-answering',
    'TEXT_CLASSIFICATION':'text-classification',
    'TEXT_GENERATION':'text-generation',
    'TEXT2TEXT_GENERATION':'text2text-generation',
    'TOKEN_CLASSIFICATION':'token-classification',
    'TRANSLATION':'translation',
    'ZERO_SHOT_CLASSIFICATION':'zero-shot-classification',
    'CONVERSATIONAL':'conversational',
    'TEXT_TO_SPEECH':'text-to-speech',
    'AUTOMATIC_SPEECH_RECOGNITION':'automatic-speech-recognition',
    'AUDIO_SOURCE_SEPERATION':'audio-source-seperation',
    'VOICE_ACTIVITY_DETECTION':'voice-activity-detection'}

# Cell
mk_class('HF_TASKS', **_hf_tasks,
        doc="A list of all HuggingFace tasks for valid API lookup as attribtues to get tab-completion and typo-proofing")

# Cell
#nbdev_comment _all_ = ['HF_TASKS']

# Internal Cell
_flair_tasks = {
    'NAMED_ENTITY_RECOGNITION':'ner',
    'PHRASE_CHUNKING':'chunk',
    'VERB_DISAMBIGUATION':'frame',
    'PART_OF_SPEECH_TAGGING':'pos',
    'UNIVERSAL_PART_OF_SPEECH_TAGGING':'upos',
    'EMBEDDINGS':'embeddings',

}

# Cell
mk_class('FLAIR_TASKS', **_flair_tasks,
        doc="A list of all Flair tasks for valid API lookup as attributes to get tab-completion and typo-proofing")

# Cell
#nbdev_comment _all_ = ['FLAIR_TASKS']

# Cell
class HFModelResult:
    """
    A very basic class for storing a HuggingFace model returned through an API request

    They have 4 properties:
      - `name`: The `modelId` from the `modelInfo`. This also includes the model author's name, such as "IlyaGusev/mbart_ru_sum_gazeta"
      - `tags`: Any tags that were included in `HuggingFace` in relation to the model.
      - `tasks`: These are the tasks dictated for the model.
    """
    def __init__(
        self,
        model_info: ModelInfo # `ModelInfo` object from HuggingFace model hub
    ):
        self.info = model_info

    def __repr__(self): return f'Model Name: {self.name}, Tasks: [' + ', '.join(self.tasks) + ']'

    @property
    def name(self): return self.info.modelId

    @property
    def tags(self): return self.info.tags

    @property
    def tasks(self):
        if self.info.pipeline_tag:
            all_tasks = [self.info.pipeline_tag]
            for tag in self.tags:
                if (tag in _hf_tasks.values()) and (tag not in all_tasks):
                    all_tasks += [tag]
        else: all_tasks = []
        all_tasks.sort()
        return all_tasks

    def to_dict(
        self
    ) -> dict: # Dictionary with keys `model_name`, `tags`, `tasks`, `model_info`
        "Returns `HFModelResult` as a dictionary"
        return {'model_name':self.name, 'tags':self.tags, 'tasks':self.tasks, 'model_info':self.info}

# Cell
class HFModelHub:
    "A class for interacting with the HF model hub API, and searching for models by name or task"

    def __init__(
        self,
        username:str=None, # Your HuggingFace username
        password:str=None # Your HuggingFace password
    ):
        self.api = HfApi()
        if username and password:
            self.token = self.api.login(username, password)
        elif username or password:
            print('Only a username or password was entered. You should include both to get authorized access')

    def _format_results(
        self,
        results:list, # A list of HuggingFace API results
        as_dict:bool=False, # Whether to return as a dictionary or list
        user_uploaded:bool=False # Whether to filter out user-uploaded results
    ) -> (List[HFModelResult], Dict[str, HFModelResult]): # A list of `HFModelResult`s
        "Takes raw HuggingFace API results and makes them easier to read and work with"
        results = apply(HFModelResult, results)
        if not user_uploaded:
            results = [r for r in results if '/' not in r.name]
        if as_dict:
            dicts = apply(Self.to_dict(), results)
            results = {m['model_name'] : m for m in dicts}
        return results

    def search_model_by_task(
        self,
        task:str, # A valid task to search in the HuggingFace hub for
        as_dict:bool=False, # Whether to return as a dictionary or list
        user_uploaded:bool=False # Whether to filter out user-uploaded results
    ) -> (List[HFModelResult], Dict[str, HFModelResult]): # A list of `HFModelResult`s
        "Searches HuggingFace Model API for all pretrained models relating to `task`"
        if task not in _hf_tasks.values():
            raise ValueError(f'''`{task}` is not a valid task.

            Please choose a valid one available from HuggingFace: (https://huggingface.co/transformers/task_summary.html)
            Or with the `HF_TASKS` object''')
        models = self.api.list_models(task)
        return self._format_results(models, as_dict, user_uploaded)

    def search_model_by_name(
        self,
        name:str, # A valid model name
        as_dict:bool=False, # Whether to return as a dictionary or list
        user_uploaded:bool=False # Whether to filter out user-uploaded results
    ) -> (List[HFModelResult], Dict[str, HFModelResult]): # A list of `HFModelResult`s
        "Searches HuggingFace Model API for all pretrained models containing `name`"
        if user_uploaded:
            models = self.api.list_models()
            models = self._format_results(models, as_dict, user_uploaded)
            models = [m for m in models if name in m.name]

        else:
            models = self.api.list_models(name)
            models = self._format_results(models, as_dict, user_uploaded)
        return models

# Internal Cell
# Flair models originating from:
# https://github.com/flairNLP/flair/blob/master/flair/models/text_classification_model.py#L483
# and: https://github.com/flairNLP/flair/blob/master/flair/models/sequence_tagger_model.py#L1053
# and: https://github.com/flairNLP/flair/blob/master/flair/embeddings/token.py#L406
_flair_models = {
    'de-offensive-language' : ['text-classification'],
    'sentiment' : ['text-classification'],
    'en-sentiment' : ['text-classification'],
    'sentiment-fast' : ['text-classification'],
    'communicative-functions' : ['text-classification'],
    'tars-base' : ['text-classification'],
    # English Named Entity Recognition Models (NER)
    'ner' : ['token-classification'],
    'ner-pooled' : ['token-classification'],
    'ner-fast' : ['token-classification'],
    'ner-ontonotes' : ['token-classification'],
    'ner-ontonotes-fast' : ['token-classification'],
    # Multilingual NER models
    'ner-multi' : ['token-classification'],
    'multi-ner' : ['token-classification'],
    'ner-multi-fast' : ['token-classification'],
    # English POS models
    'upos' : ['token-classification'],
    'upos-fast' : ['token-classification'],
    'pos' : ['token-classification'],
    'pos-fast' : ['token-classification'],
    # Multilingual POS models
    'pos-multi' : ['token-classification'],
    'multi-pos' : ['token-classification'],
    'pos-multi-fast' : ['token-classification'],
    'multi-pos-fast' : ['token-classification'],
    # English SRL models
    'frame' : ['token-classification'],
    'frame-fast' : ['token-classification'],
    # English chunking models
    'chunk' : ['token-classification'],
    'chunk-fast' : ['token-classification'],
    # Danish models
    'da-pos' : ['token-classification'],
    'da-ner' : ['token-classification'],
    # German models
    'de-pos' : ['token-classification'],
    'de-pos-tweets' : ['token-classification'],
    'de-ner' : ['token-classification'],
    'de-ner-germeval' : ['token-classification'],
    'de-ler' : ['token-classification'],
    'de-ner-legal' : ['token-classification'],
    # French models
    'fr-ner' : ['token-classification'],
    # Dutch models
    'nl-ner' : ['token-classification'],
    'nl-ner-rnn' : ['token-classification'],
    # Malayalam models
    'ml-pos' : ['token-classification'],
    'ml-upos' : ['token-classification'],
    # Portuguese models
    'pt-pos-clinical' : ['token-classification'],
    # Keyphrase models
    'keyphrase' : ['token-classification'],
    'negation-speculation' : ['token-classification'],
    # Biomedical
    'hunflair-paper-cellline' : ['token-classification'],
    'hunflair-paper-chemical' : ['token-classification'],
    'hunflair-paper-disease' : ['token-classification'],
    'hunflair-paper-gene' : ['token-classification'],
    'hunflair-paper-species' : ['token-classification'],
    'hunflair-cellline' : ['token-classification'],
    'hunflair-chemical' : ['token-classification'],
    'hunflair-disease' : ['token-classification'],
    'hunflair-gene' : ['token-classification'],
    'hunflair-species' : ['token-classification'],
    # Embeddings
    # multilingual models
    "multi-forward":['embeddings'],
    "multi-backward":['embeddings'],
    "multi-v0-forward":['embeddings'],
    "multi-v0-backward":['embeddings'],
    "multi-forward-fast":['embeddings'],
    "multi-backward-fast":['embeddings'],
    # English models
    "en-forward":['embeddings'],
    "en-backward":['embeddings'],
    "en-forward-fast":['embeddings'],
    "en-backward-fast":['embeddings'],
    "news-forward":['embeddings'],
    "news-backward":['embeddings'],
    "news-forward-fast":['embeddings'],
    "news-backward-fast":['embeddings'],
    "mix-forward":['embeddings'],
    "mix-backward":['embeddings'],
    # Arabic
    "ar-forward":['embeddings'],
    "ar-backward":['embeddings'],
    # Bulgarian
    "bg-forward-fast":['embeddings'],
    "bg-backward-fast":['embeddings'],
    "bg-forward":['embeddings'],
    "bg-backward":['embeddings'],
    # Czech
    "cs-forward":['embeddings'],
    "cs-backward":['embeddings'],
    "cs-v0-forward":['embeddings'],
    "cs-v0-backward":['embeddings'],
    # Danish
    "da-forward":['embeddings'],
    "da-backward":['embeddings'],
    # German
    "de-forward":['embeddings'],
    "de-backward":['embeddings'],
    "de-historic-ha-forward":['embeddings'],
    "de-historic-ha-backward":['embeddings'],
    "de-historic-wz-forward":['embeddings'],
    "de-historic-wz-backward":['embeddings'],
    "de-historic-rw-forward":['embeddings'],
    "de-historic-rw-backward":['embeddings'],
    # Spanish
    "es-forward":['embeddings'],
    "es-backward":['embeddings'],
    "es-forward-fast":['embeddings'],
    "es-backward-fast":['embeddings'],
    # Basque
    "eu-forward":['embeddings'],
    "eu-backward":['embeddings'],
    "eu-v1-forward":['embeddings'],
    "eu-v1-backward":['embeddings'],
    "eu-v0-forward":['embeddings'],
    "eu-v0-backward":['embeddings'],
    # Persian
    "fa-forward":['embeddings'],
    "fa-backward":['embeddings'],
    # Finnish
    "fi-forward":['embeddings'],
    "fi-backward":['embeddings'],
    # French
    "fr-forward":['embeddings'],
    "fr-backward":['embeddings'],
    # Hebrew
    "he-forward":['embeddings'],
    "he-backward":['embeddings'],
    # Hindi
    "hi-forward":['embeddings'],
    "hi-backward":['embeddings'],
    # Croatian
    "hr-forward":['embeddings'],
    "hr-backward":['embeddings'],
    # Indonesian
    "id-forward":['embeddings'],
    "id-backward":['embeddings'],
    # Italian
    "it-forward":['embeddings'],
    "it-backward":['embeddings'],
    # Japanese
    "ja-forward":['embeddings'],
    "ja-backward":['embeddings'],
    # Malayalam
    "ml-forward":['embeddings'],
    "ml-backward":['embeddings'],
    # Dutch
    "nl-forward":['embeddings'],
    "nl-backward":['embeddings'],
    "nl-v0-forward":['embeddings'],
    "nl-v0-backward":['embeddings'],
    # Norwegian
    "no-forward":['embeddings'],
    "no-backward":['embeddings'],
    # Polish
    "pl-forward":['embeddings'],
    "pl-backward":['embeddings'],
    "pl-opus-forward":['embeddings'],
    "pl-opus-backward":['embeddings'],
    # Portuguese
    "pt-forward":['embeddings'],
    "pt-backward":['embeddings'],
    # Pubmed
    "pubmed-forward":['embeddings'],
    "pubmed-backward":['embeddings'],
    "pubmed-2015-forward":['embeddings'],
    "pubmed-2015-backward":['embeddings'],
    # Slovenian
    "sl-forward":['embeddings'],
    "sl-backward":['embeddings'],
    "sl-v0-forward":['embeddings'],
    "sl-v0-backward":['embeddings'],
    # Swedish
    "sv-forward":['embeddings'],
    "sv-backward":['embeddings'],
    "sv-v0-forward":['embeddings'],
    "sv-v0-backward":['embeddings'],
    # Tamil
    "ta-forward":['embeddings'],
    "ta-backward":['embeddings'],
    # CLEF HIPE Shared task
    "de-impresso-hipe-v1-forward":['embeddings'],
    "de-impresso-hipe-v1-backward":['embeddings'],
    "en-impresso-hipe-v1-forward":['embeddings'],
    "en-impresso-hipe-v1-backward":['embeddings'],
    "fr-impresso-hipe-v1-forward":['embeddings'],
    "fr-impresso-hipe-v1-backward":['embeddings']
}

# Cell
FLAIR_MODELS = [ModelInfo(f'flairNLP/{key}', pipeline_tag=val[0]) for key,val in _flair_models.items()]

# Cell
class FlairModelResult(HFModelResult):
    """
    A version of `HFModelResult` for Flair specifically.

    Includes which backend the model was found (such as on HuggingFace or Flair's private model list)
    """

    def __init__(
        self,
        model_info: ModelInfo # ModelInfo object from HuggingFace model hub
    ):
        if 'flairNLP' in model_info.modelId:
            self.from_hf = False
        else:
            self.from_hf = True
        super().__init__(model_info)

    def __repr__(self): return f'Model Name: {self.name.replace("flairNLP", "flair")}, Tasks: [' + ', '.join(self.tasks) + ']' + f', Source: {self.source}'

    @property
    def source(self):
        if self.from_hf: return "HuggingFace Model Hub"
        else: return "Flair's Private Model Hub"

# Cell
class FlairModelHub:
    "A class for interacting with the HF model hub API, and searching for Flair models by name or task"

    def __init__(
        self,
        username:str=None, # HuggingFace username
        password:str=None # HuggingFace password
    ):
        self.api = HfApi()
        if username and password:
            self.token = self.api.login(username, password)
        elif username or password:
            print('Only a username or password was entered. You should include both to get authorized access')
        self.models = self.api.list_models('flair') + FLAIR_MODELS

    def _format_results(
        self,
        results:list, # A list of HuggingFace API results
        as_dict:bool=False, # Whether to return as a dictionary or list
        user_uploaded:bool=False # Whether to filter out user-uploaded results
    ) -> (List[FlairModelResult], Dict[str, FlairModelResult]): # A list of `FlairModelResult`s
        "Takes raw HuggingFace API results and makes them easier to read and work with"
        results = apply(FlairModelResult, results)
        if not user_uploaded:
            results = [r for r in results if 'flair/' in r.name or 'flairNLP/' in r.name]
        if as_dict:
            dicts = apply(Self.to_dict(), results)
            results = {m['model_name'] : m for m in dicts}
        return results

    def search_model_by_name(
        self,
        name:str, # A valid model name
        as_dict:bool=False, # Whether to return as a dictionary or list
        user_uploaded:bool=False # Whether to filter out user-uploaded results
    ) -> (List[FlairModelResult], Dict[str, FlairModelResult]): # A list of `FlairModelResult`s
        "Searches HuggingFace Model API for all flair models containing `name`"
        models = [m for m in self.models if name in m.modelId]
        return self._format_results(models, as_dict, user_uploaded)

    def search_model_by_task(
        self,
        task:str, # A valid task to search the HuggingFace hub for
        as_dict=False, # Whether to return as a dictionary or list
        user_uploaded=False # Whether to filter out user-uploaded results
    ) -> (List[FlairModelResult], Dict[str, FlairModelResult]): # A list of `FlairModelResult`s
        "Searches HuggingFace Model API for all flair models for `task`"
        if (task not in _flair_tasks.values()) and (task != ''):
            raise ValueError(f'''`{task}` is not a valid task.

            Please choose a valid one available from Flair: (https://huggingface.co/flair)
            Or with the `FLAIR_TASKS` object''')
        models = [m for m in self.models if task in m.modelId or task == m.pipeline_tag]
        return self._format_results(models, as_dict, user_uploaded)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/13_training.sequence_classification.ipynb (unless otherwise specified).

__all__ = ['SequenceClassificationDatasets', 'SequenceClassificationTuner']

# Cell
import pandas as pd
from fastcore.foundation import L
from fastcore.meta import delegates
from fastcore.xtras import Path, range_of

from fastai.basics import *
from fastai.data.transforms import get_files

from datasets import Dataset
from transformers import AutoModelForSequenceClassification, default_data_collator, AutoTokenizer

from .core import * # Core has everything we need so you should always import * with it

from ..inference.sequence_classification import TransformersSequenceClassifier, SequenceResult, DetailLevel
from typing import List

from .arrow_utils import TextNoNewLineDatasetReader

# Cell
def _tokenize(item, tokenizer, tokenize_kwargs): return tokenizer(item['text'], **tokenize_kwargs)

# Cell
class SequenceClassificationDatasets(TaskDatasets):
    """
    A set of datasets designed for sequence classification
    """
    def __init__(
        self,
        train_dset:Dataset, # A training dataset
        valid_dset:Dataset, # A validation dataset
        tokenizer_name:str, # The name of a tokenizer
        tokenize:bool, # Whether to tokenize immediatly
        tokenize_kwargs:dict, # kwargs for the tokenize function
        auto_kwargs:dict, # AutoTokenizer.from_pretrained kwargs
        remove_columns:list, # The columns to remove when tokenizing
        categorize:[Categorize,MultiCategorize] # A Categorize instance
    ):
        "Constructs TaskDatasets, should not be called explicitly"
        super().__init__(
            train_dset,
            valid_dset,
            tokenizer_name,
            tokenize,
            _tokenize,
            tokenize_kwargs,
            auto_kwargs,
            remove_columns
        )
        self.categorize = categorize


    @classmethod
    def from_dfs(
        cls,
        train_df:pd.DataFrame, # A training dataframe
        text_col:str, # The name of the text column
        label_col:str, # The name of the label column
        tokenizer_name:str, # The name of the tokenizer
        tokenize:bool=True, # Whether to tokenize immediatly
        is_multicategory:bool=False, # Whether each item has a single label or multiple labels
        label_delim=' ', # If `is_multicategory`, how to separate the labels
        valid_df=None, # An optional validation dataframe
        split_func=None, # Optionally a splitting function similar to RandomSplitter
        split_pct=.2, # What % to split the train_df
        tokenize_kwargs:dict={}, # kwargs for the tokenize function
        auto_kwargs:dict={} # kwargs for the AutoTokenizer.from_pretrained constructor
    ):
        "Builds `SequenceClassificationDatasets` from a `DataFrame` or set of `DataFrames`"
        if split_func is None: split_func = RandomSplitter(split_pct)
        if valid_df is None:
            train_idxs, valid_idxs = split_func(train_df)
            valid_df = train_df.iloc[valid_idxs]
            train_df = train_df.iloc[train_idxs]

        train_df = train_df[[text_col,label_col]]
        valid_df = valid_df[[text_col,label_col]]
        train_df = train_df.rename(columns={text_col:'text', label_col: 'label'})
        valid_df = valid_df.rename(columns={text_col:'text', label_col: 'label'})

        lbls = list(train_df['label'].unique())
        if is_multicategory:
            classes = set()
            for lbl in lbls:
                sep_l = lbl.split(label_delim)
                for l in sep_l: classes.add(l)
            categorize = MultiCategorize(classes)
        else:
            classes = set()
            for lbl in lbls: classes.add(lbl)
            categorize = Categorize(classes)

        train_df['label'] = train_df['label'].apply(lambda x: categorize(x))
        valid_df['label'] = valid_df['label'].apply(lambda x: categorize(x))


        train_dset = Dataset.from_dict(train_df.to_dict('list'))
        valid_dset = Dataset.from_dict(valid_df.to_dict('list'))
        return cls(
            train_dset,
            valid_dset,
            tokenizer_name,
            tokenize,
            tokenize_kwargs,
            auto_kwargs,
            remove_columns=['text'],
            categorize=categorize
        )

    @classmethod
    def from_csvs(
        cls,
        train_csv:Path, # A training csv file
        text_col:str, # The name of the text column
        label_col:str, # The name of the label column
        tokenizer_name:str, # The name of the tokenizer
        tokenize:bool=True, # Whether to tokenize immediatly
        is_multicategory:bool=False, # Whether each item has a single label or multiple labels
        label_delim=' ', # If `is_multicategory`, how to separate the labels
        valid_csv:Path=None, # An optional validation csv
        split_func=None, # Optionally a splitting function similar to RandomSplitter
        split_pct=.2, # What % to split the train_df
        tokenize_kwargs:dict={}, # kwargs for the tokenize function
        auto_kwargs:dict={}, # kwargs for the AutoTokenizer.from_pretrained constructor
        **kwargs, # kwargs for `pd.read_csv`
    ):
        "Builds `SequenceClassificationDatasets` from a single csv or set of csvs. A convience constructor for `from_dfs`"
        train_df = pd.read_csv(train_csv, **kwargs)
        if valid_csv is not None: valid_df = pd.read_csv(valid_csv, **kwargs)
        else: valid_df = None
        return cls.from_dfs(
            train_df,
            text_col,
            label_col,
            tokenizer_name,
            tokenize,
            is_multicategory,
            label_delim,
            valid_df,
            split_func,
            split_pct,
            tokenize_kwargs,
            auto_kwargs
        )

    @classmethod
    def from_folders(
        cls,
        train_path:Path, # The path to the training data
        get_label:callable, # A function which grabs the label(s) given a text files `Path`
        tokenizer_name:str, # The name of the tokenizer
        tokenize:bool=True, # Whether to tokenize immediatly
        is_multicategory:bool=False, # Whether each item has a single label or multiple labels
        label_delim='_', # if `is_multicategory`, how to separate the labels
        valid_path:Path=None, # The path to the validation data
        split_func=None, # Optionally a splitting function similar to RandomSplitter
        split_pct=.2, # What % to split the items in the `train_path`
        tokenize_kwargs:dict={}, # kwargs for the tokenize function
        auto_kwargs:dict={}, # kwargs for the AutoTokenizer.from_pretrained constructor
    ):
        "Builds `SequenceClassificationDatasets` from a folder or groups of folders"
        train_txts = get_files(train_path, extensions='.txt')
        if valid_path is not None:
            valid_txts = get_files(valid_path, extensions='.txt')
        else:
            if split_func is None:
                split_func = RandomSplitter(split_pct)
            train_idxs, valid_idxs = split_func(train_txts)
            valid_txts = train_txts[valid_idxs]
            train_txts = train_txts[train_idxs]
        train_txts = [str(x) for x in train_txts]
        valid_txts = [str(x) for x in valid_txts]
        train_dset = TextNoNewLineDatasetReader(train_txts).read()
        valid_dset = TextNoNewLineDatasetReader(valid_txts).read()

        train_lbls = [get_label(o) for o in train_txts]
        valid_lbls = [get_label(o) for o in valid_txts]
        if is_multicategory:
            classes = set()
            for lbl in train_lbls:
                sep_l = lbl.split(label_delim)
                for l in sep_l: classes.add(l)
            categorize = MultiCategorize(classes)
        else:
            classes = set()
            for lbl in train_lbls: classes.add(lbl)
            categorize = Categorize(classes)
        train_lbls = [categorize(lbl) for lbl in train_lbls]
        valid_lbls = [categorize(lbl) for lbl in valid_lbls]
        train_dset = train_dset.add_column('label', train_lbls)
        valid_dset = valid_dset.add_column('label', valid_lbls)

        return cls(
            train_dset,
            valid_dset,
            tokenizer_name,
            tokenize,
            tokenize_kwargs,
            auto_kwargs,
            remove_columns=['text'],
            categorize=categorize
        )

    @delegates(DataLoaders)
    def dataloaders(
        self,
        batch_size=8, # A batch size
        shuffle_train=True, # Whether to shuffle the training dataset
        collate_fn = None, # A custom collation function
        **kwargs, # Torch DataLoader kwargs
    ):
        "Build DataLoaders from `self`"
        dls = super().dataloaders(batch_size, shuffle_train, collate_fn, **kwargs)
        dls[0].categorize = self.categorize
        return dls

# Cell
class SequenceClassificationTuner(AdaptiveTuner):
    """
    An `AdaptiveTuner` with good defaults for Sequence Classification tasks

    **Valid kwargs and defaults:**
      - `lr`:float = 0.001
      - `splitter`:function = `trainable_params`
      - `cbs`:list = None
      - `path`:Path = None
      - `model_dir`:Path = 'models'
      - `wd`:float = None
      - `wd_bn_bias`:bool = False
      - `train_bn`:bool = True
      - `moms`: tuple(float) = (0.95, 0.85, 0.95)

    """
    def __init__(
        self,
        dls:DataLoaders, # A set of DataLoaders
        model_name:str, # A HuggingFace model
        tokenizer = None, # A HuggingFace tokenizer
        loss_func = CrossEntropyLossFlat(), # A loss function
        metrics = [accuracy, F1Score()], # Metrics to monitor the training with
        opt_func = Adam, # A fastai or torch Optimizer
        additional_cbs = None, # Additional Callbacks to have always tied to the Tuner,
        expose_fastai_api = False, # Whether to expose the fastai API
        num_classes:int=None, # The number of classes
        **kwargs, # kwargs for `Learner.__init__`
    ):
        additional_cbs = listify(additional_cbs)
        for arg in 'dls,model,loss_func,metrics,opt_func,cbs,expose_fastai'.split(','):
            if arg in kwargs.keys(): kwargs.pop(arg) # Pop all existing kwargs
        if hasattr(dls[0], 'categorize'): num_classes = getattr(dls[0].categorize, 'classes', None)
        if num_classes is None: raise ValueError("Could not extrapolate number of classes, please pass it in as a param")
        if not isinstance(num_classes, int): num_classes = len(num_classes)
        model = AutoModelForSequenceClassification.from_pretrained(model_name, num_labels=num_classes)
        if tokenizer is None: tokenizer = AutoTokenizer.from_pretrained(model_name)

        super().__init__(
            expose_fastai_api,
            dls = dls,
            model = model,
            tokenizer = tokenizer,
            loss_func = loss_func,
            metrics = metrics,
            opt_func = opt_func,
            cbs=additional_cbs,
            **kwargs
        )

    def predict(
        self,
        text:Union[List[str], str], # Some text or list of texts to do inference with
        bs:int=64, # A batch size to use for multiple texts
        detail_level:DetailLevel = DetailLevel.Low, # A detail level to return on the predictions
        class_names:list = None, # A list of labels
    ) -> dict: # A dictionary of filtered predictions
        "Predict some `text` for sequence classification with the currently loaded model"
        if getattr(self, '_inferencer', None) is None: self._inferencer = TransformersSequenceClassifier(self.tokenizer, self.model)
        if class_names is None:
            preds = self._inferencer.predict(text,bs)
            cat = getattr(self.dls, 'categorize', None)
            vocab = cat.classes if cat is not None else None
        else:
            preds = self._inferencer.predict(text,bs,class_names=class_names)
            vocab = class_names
        return SequenceResult(preds, vocab).to_dict(detail_level)


# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"SentenceResult": "00_result.ipynb",
         "GatherInputsCallback": "01_callback.ipynb",
         "SetInputsCallback": "01_callback.ipynb",
         "GeneratorCallback": "01_callback.ipynb",
         "HFModelResult": "02_model_hub.ipynb",
         "HFModelHub": "02_model_hub.ipynb",
         "FLAIR_MODELS": "02_model_hub.ipynb",
         "FlairModelResult": "02_model_hub.ipynb",
         "FlairModelHub": "02_model_hub.ipynb",
         "DataLoader.one_batch": "03_model.ipynb",
         "GatherPredsCallback.after_validate": "03_model.ipynb",
         "CudaCallback": "03_model.ipynb",
         "AdaptiveModel": "03_model.ipynb",
         "logger": "11_inference.utils.ipynb",
         "EmbeddingResult": "04_embeddings.ipynb",
         "EasyWordEmbeddings": "04_embeddings.ipynb",
         "EasyStackedEmbeddings": "04_embeddings.ipynb",
         "EasyDocumentEmbeddings": "04_embeddings.ipynb",
         "TokenClassificationResult": "05_token_classification.ipynb",
         "TransformersTokenTagger": "05_token_classification.ipynb",
         "FlairTokenTagger": "05_token_classification.ipynb",
         "EasyTokenTagger": "05_token_classification.ipynb",
         "SequenceResult": "06_sequence_classification.ipynb",
         "TransformersSequenceClassifier": "06_sequence_classification.ipynb",
         "FlairSequenceClassifier": "06_sequence_classification.ipynb",
         "EasySequenceClassifier": "06_sequence_classification.ipynb",
         "TransformersSummarizer": "07_summarization.ipynb",
         "EasySummarizer": "07_summarization.ipynb",
         "TranslationResult": "08_translation.ipynb",
         "TransformersTranslator": "08_translation.ipynb",
         "EasyTranslator": "08_translation.ipynb",
         "TransformersTextGenerator": "09_text_generation.ipynb",
         "EasyTextGenerator": "09_text_generation.ipynb",
         "QACallback": "10_question_answering.ipynb",
         "QAResult": "10_question_answering.ipynb",
         "TransformersQuestionAnswering": "10_question_answering.ipynb",
         "EasyQuestionAnswering": "10_question_answering.ipynb",
         "normalize_answer": "11_inference.utils.ipynb",
         "get_tokens": "11_inference.utils.ipynb",
         "compute_exact": "11_inference.utils.ipynb",
         "compute_f1": "11_inference.utils.ipynb",
         "get_raw_scores": "11_inference.utils.ipynb",
         "apply_no_ans_threshold": "11_inference.utils.ipynb",
         "make_eval_dict": "11_inference.utils.ipynb",
         "merge_eval": "11_inference.utils.ipynb",
         "find_best_thresh_v2": "11_inference.utils.ipynb",
         "find_all_best_thresh_v2": "11_inference.utils.ipynb",
         "find_best_thresh": "11_inference.utils.ipynb",
         "find_all_best_thresh": "11_inference.utils.ipynb",
         "squad_evaluate": "11_inference.utils.ipynb",
         "get_final_text": "11_inference.utils.ipynb",
         "compute_predictions_logits": "11_inference.utils.ipynb",
         "compute_predictions_log_probs": "11_inference.utils.ipynb",
         "ParentLabeller": "12_training.core.ipynb",
         "ColReader": "12_training.core.ipynb",
         "Categorize": "12_training.core.ipynb",
         "MultiCategorize": "12_training.core.ipynb",
         "RandomSplitter": "12_training.core.ipynb",
         "TaskDatasets": "12_training.core.ipynb",
         "AdaptiveDataLoaders": "12_training.core.ipynb",
         "AdaptiveTuner": "12_training.core.ipynb",
         "SequenceClassificationDatasets": "13_training.sequence_classification.ipynb",
         "SequenceClassificationTuner": "13_training.sequence_classification.ipynb",
         "LanguageModelDatasets": "14_training.language_model.ipynb",
         "LanguageModelTuner": "14_training.language_model.ipynb",
         "TextNoNewLine": "15_training.arrow_utils.ipynb",
         "TextNoNewLineDatasetReader": "15_training.arrow_utils.ipynb"}

modules = ["result.py",
           "callback.py",
           "model_hub.py",
           "model.py",
           "inference/embeddings.py",
           "inference/token_classification.py",
           "inference/sequence_classification.py",
           "inference/summarization.py",
           "inference/translation.py",
           "inference/text_generation.py",
           "inference/question_answering.py",
           "inference/utils.py",
           "training/core.py",
           "training/sequence_classification.py",
           "training/language_model.py",
           "training/arrow_utils.py"]

doc_url = "https://novetta.github.io/adaptnlp/"

git_url = "https://github.com/novetta/adaptnlp/tree/master/"

def custom_doc_links(name): return None
